// Copyright 2019 TiKV Project Authors. Licensed under Apache-2.0.

use std::sync::Arc;

use criterion::black_box;

use tipb::expression::Expr;

use tikv::coprocessor::dag::batch::executors::BatchTopNExecutor;
use tikv::coprocessor::dag::executor::{Executor, TopNExecutor};
use tikv::coprocessor::dag::expr::EvalConfig;

use crate::util::bencher::Bencher;
use crate::util::executor_descriptor::top_n;
use crate::util::FixtureBuilder;

pub trait TopNBencher {
    fn name(&self) -> &'static str;

    fn bench(
        &self,
        b: &mut criterion::Bencher,
        fb: &FixtureBuilder,
        order_by_expr: &[Expr],
        order_is_desc: &[bool],
        n: usize,
    );

    fn box_clone(&self) -> Box<dyn TopNBencher>;
}

impl Clone for Box<dyn TopNBencher> {
    #[inline]
    fn clone(&self) -> Self {
        self.box_clone()
    }
}

/// A bencher that will use normal top N executor to bench the giving aggregate
/// expression.
pub struct NormalBencher;

impl TopNBencher for NormalBencher {
    fn name(&self) -> &'static str {
        "normal"
    }

    fn bench(
        &self,
        b: &mut criterion::Bencher,
        fb: &FixtureBuilder,
        order_by_expr: &[Expr],
        order_is_desc: &[bool],
        n: usize,
    ) {
        crate::util::bencher::NormalNextAllBencher::new(|| {
            assert_eq!(order_by_expr.len(), order_is_desc.len());
            let meta = top_n(order_by_expr, order_is_desc, n).take_topN();
            let src = fb.clone().build_normal_fixture_executor();
            Box::new(
                TopNExecutor::new(
                    black_box(meta),
                    black_box(Arc::new(EvalConfig::default())),
                    black_box(Box::new(src)),
                )
                .unwrap(),
            ) as Box<dyn Executor>
        })
        .bench(b);
    }

    fn box_clone(&self) -> Box<dyn TopNBencher> {
        Box::new(Self)
    }
}

/// A bencher that will use batch top N executor to bench the giving aggregate
/// expression.
pub struct BatchBencher;

impl TopNBencher for BatchBencher {
    fn name(&self) -> &'static str {
        "batch"
    }

    fn bench(
        &self,
        b: &mut criterion::Bencher,
        fb: &FixtureBuilder,
        order_by_expr: &[Expr],
        order_is_desc: &[bool],
        n: usize,
    ) {
        crate::util::bencher::BatchNextAllBencher::new(|| {
            let src = fb.clone().build_batch_fixture_executor();
            Box::new(
                BatchTopNExecutor::new(
                    black_box(Arc::new(EvalConfig::default())),
                    black_box(Box::new(src)),
                    black_box(order_by_expr.to_vec()),
                    black_box(order_is_desc.to_vec()),
                    black_box(n),
                )
                .unwrap(),
            )
        })
        .bench(b);
    }

    fn box_clone(&self) -> Box<dyn TopNBencher> {
        Box::new(Self)
    }
}
